# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create an ServiceHotWater object that can be used to specify hot water usage in
a ProgramType.
-

    Args:
        _name_: Text to set the name for the ServiceHotWater and to be incorporated
            into a unique ServiceHotWater identifier. If None, a unique name will
            be generated.
        _flow_per_area: A numerical value for the total volume flow rate of water
            per unit area of floor (L/h-m2).
        _schedule: A fractional schedule for the use of hot water over the course of
            the year. The fractional values will get multiplied by the
            _flow_per_area to yield a complete water usage profile.
        _target_temp_: The target temperature of the water out of the tap in Celsius.
            This the temperature after the hot water has been mixed with cold
            water from the water mains. The default value assumes that the
            flow_per_area on this object is only for water straight out
            of the water heater. (Default: 60C).
        _sensible_fract_: A number between 0 and 1 for the fraction of the total
            hot water load given off as sensible heat in the zone. (Default: 0.2).
        _latent_fract_: A number between 0 and 1 for the fraction of the total
            hot water load that is latent (as opposed to sensible). (Default: 0.05).

    Returns:
        hot_water: A ServiceHotWater object that can be used to specify hot water usage
            in a ProgramType.
"""

ghenv.Component.Name = 'HB Service Hot Water'
ghenv.Component.NickName = 'ServiceHotWater'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.hotwater import ServiceHotWater
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # make a default ServiceHotWater name if none is provided
    name = clean_and_id_ep_string('ServiceHotWater') if _name_ is None else \
        clean_ep_string(_name_)

    # get the schedule
    if isinstance(_schedule, str):
        _schedule = schedule_by_identifier(_schedule)

    # get default radiant, latent, and lost fractions
    _target_temp_ = _target_temp_ if _target_temp_ is not None else 60
    _sensible_fract_ = _sensible_fract_ if _sensible_fract_ is not None else 0.2
    _latent_fract_ = _latent_fract_ if _latent_fract_ is not None else 0.05

    # create the ServiceHotWater object
    hot_water = ServiceHotWater(name, _flow_per_area, _schedule, _target_temp_,
                                _sensible_fract_, _latent_fract_)
    if _name_ is not None:
        hot_water.display_name = _name_


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a Lighting object that can be used to create a ProgramType or be assigned
directly to a Room.
-

    Args:
        _name_: Text to set the name for the Lighting and to be incorporated
            into a unique Lighting identifier. If None, a unique name will
            be generated.
         _watts_per_area: A numerical value for the lighting power density in
            Watts per square meter of floor area.
        _schedule: A fractional for the use of lights over the course of the year.
            The fractional values will get multiplied by the _watts_per_area to
            yield a complete lighting profile.
        _radiant_fract_: A number between 0 and 1 for the fraction of the total
            lighting load given off as long wave radiant heat.
            Default: 0.32 (representative of pendant lighting).
        _visible_fract_: A number between 0 and 1 for the fraction of the total
            lighting load given off as short wave visible light.
            Default: 0.25  (representative of pendant lighting).
        return_fract_: A number between 0 and 1 for the fraction of the total
            lighting load that goes into the zone return air (into the zone outlet
            node). Default: 0.0 (representative of pendant lighting).
        baseline_: An optional number for the baseline lighting power density in
            W/m2 of floor area. This baseline is useful to track how much
            better the installed lights are in comparison to a standard like
            ASHRAE 90.1. If set to None, it will default to 11.84029 W/m2,
            which is that ASHRAE 90.1-2004 baseline for an office.

    Returns:
        lighting: A Lighting object that can be used to create a ProgramType or
            be assigned directly to a Room.
"""

ghenv.Component.Name = "HB Lighting"
ghenv.Component.NickName = 'Lighting'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "3"

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.lighting import Lighting
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # make a default Lighting name if none is provided
    name = clean_and_id_ep_string('Lighting') if _name_ is None else \
        clean_ep_string(_name_)

    # get the schedule
    if isinstance(_schedule, str):
        _schedule = schedule_by_identifier(_schedule)

    # get default radiant, visible, and return fractions
    return_fract_ = return_fract_ if return_fract_ is not None else 0.0
    _radiant_fract_ = _radiant_fract_ if _radiant_fract_ is not None else 0.32
    _visible_fract_ = _visible_fract_ if _visible_fract_ is not None else 0.25

    # create the Lighting object
    lighting = Lighting(name, _watts_per_area, _schedule,
                        return_fract_, _radiant_fract_, _visible_fract_)
    if _name_ is not None:
        lighting.display_name = _name_
    if baseline_ is not None:
        lighting.baseline_watts_per_area = baseline_


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a list of ground constructions that can be used to edit or create a
ConstructionSet object.
-

    Args:
        _ground_wall_: A construction object for underground walls (or text for
            the identifier of the construction within the library).
        _ground_roof_: A construction object for underground roofs (or text for
            the identifier of the construction within the library).
        _ground_floor_: A construction object for ground-contact floors (or text
            for the identifier of the construction within the library).
    
    Returns:
        ground_set: A list of ground constructions that can be used to edit
            or create a ConstructionSet object.
"""

ghenv.Component.Name = "HB Ground Construction Subset"
ghenv.Component.NickName = 'GroundSubset'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = "4"

try:  # import honeybee_energy dependencies
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


def opaque_constr(construction, input_name):
    """Get an OpaqueConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return opaque_construction_by_identifier(construction)
    else:
        assert isinstance(construction, OpaqueConstruction), \
            'Expected OpaqueConstruction for {}. Got {}'.format(
                input_name, type(construction))
    return construction


# go through each input construction
if _ground_wall_ is not None:
    _ground_wall_ = opaque_constr(_ground_wall_, '_ground_wall_')
if _ground_roof_ is not None:
    _ground_roof_ = opaque_constr(_ground_roof_, '_ground_roof_')
if _ground_floor_ is not None:
    _ground_floor_ = opaque_constr(_ground_floor_, '_ground_floor_')


# return the final list from the component
ground_set = [_ground_wall_, _ground_roof_, _ground_floor_]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse all of the common Room-level energy-related results from an SQL result file
that has been generated from an energy simulation.

-
    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.

    Returns:
        cooling: DataCollections for the cooling energy in kWh. For Ideal Air
            loads, this output is the sum of sensible and latent heat that must
            be removed from each room.  For detailed HVAC systems, this output
            will be electric energy needed to power each chiller/cooling coil.
        heating: DataCollections for the heating energy needed in kWh. For Ideal
            Air loads, this is the heat that must be added to each room.  For
            detailed HVAC systems, this will be fuel energy or electric energy
            needed for each boiler/heating element.
        lighting: DataCollections for the electric lighting energy used for
            each room in kWh.
        electric_equip: DataCollections for the electric equipment energy used
            for each room in kWh.
        gas_equip: DataCollections for the gas equipment energy used for each
            room in kWh.
        process: DataCollections for the process load energy used for each
            room in kWh.
        hot_water: DataCollections for the service hote water energy used for each
            room in kWh.
        fan_electric: DataCollections for the fan electric energy in kWh for
            either a ventilation fan or a HVAC system fan.
        pump_electric: DataCollections for the water pump electric energy in kWh
            for a heating/cooling system.
        people_gain: DataCollections for the internal heat gains in each room
            resulting from people (kWh).
        solar_gain: DataCollections for the total solar gain in each room (kWh).
        infiltration_load: DataCollections for the heat loss (negative) or heat
            gain (positive) in each room resulting from infiltration (kWh).
        mech_vent_load: DataCollections for the heat loss (negative) or heat gain
            (positive) in each room resulting from the outdoor air coming through
            the HVAC System (kWh).
        nat_vent_load: DataCollections for the heat loss (negative) or heat gain
            (positive) in each room resulting from natural ventilation (kWh).
"""

ghenv.Component.Name = 'HB Read Room Energy Result'
ghenv.Component.NickName = 'RoomEnergyResult'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess
import json

try:
    from ladybug.sql import SQLiteResult
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.loadbalance import LoadBalance
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def subtract_loss_from_gain(gain_load, loss_load):
    """Create a single DataCollection from gains and losses."""
    total_loads = []
    for gain, loss in zip(gain_load, loss_load):
        total_load = gain - loss
        total_load.header.metadata['type'] = \
            total_load.header.metadata['type'].replace('Gain ', '')
        total_loads.append(total_load)
    return total_loads


def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    if len(data_dicts) == 0:
        return []
    elif data_dicts[0]['type'] == 'HourlyContinuous':
        return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Monthly':
        return [MonthlyCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Daily':
        return [DailyCollection.from_dict(data) for data in data_dicts]


# List of all the output strings that will be requested
cooling_outputs = LoadBalance.COOLING + (
    'Cooling Coil Electricity Energy',
    'Chiller Electricity Energy',
    'Zone VRF Air Terminal Cooling Electricity Energy',
    'VRF Heat Pump Cooling Electricity Energy',
    'Chiller Heater System Cooling Electricity Energy',
    'District Cooling Chilled Water Energy',
    'Evaporative Cooler Electricity Energy')
heating_outputs = LoadBalance.HEATING + (
    'Boiler NaturalGas Energy',
    'Heating Coil Total Heating Energy',
    'Heating Coil NaturalGas Energy',
    'Heating Coil Electricity Energy',
    'Humidifier Electricity Energy',
    'Zone VRF Air Terminal Heating Electricity Energy',
    'VRF Heat Pump Heating Electricity Energy',
    'VRF Heat Pump Defrost Electricity Energy',
    'VRF Heat Pump Crankcase Heater Electricity Energy',
    'Chiller Heater System Heating Electricity Energy',
    'District Heating Hot Water Energy',
    'Baseboard Electricity Energy',
    'Hot_Water_Loop_Central_Air_Source_Heat_Pump Electricity Consumption',
    'Boiler Electricity Energy',
    'Water Heater NaturalGas Energy',
    'Water Heater Electricity Energy',
    'Cooling Coil Water Heating Electricity Energy')
lighting_outputs = LoadBalance.LIGHTING
electric_equip_outputs = LoadBalance.ELECTRIC_EQUIP
gas_equip_outputs = LoadBalance.GAS_EQUIP
process_outputs = LoadBalance.PROCESS
shw_outputs = ('Water Use Equipment Heating Energy',) + LoadBalance.HOT_WATER
fan_electric_outputs = (
    'Zone Ventilation Fan Electricity Energy',
    'Fan Electricity Energy',
    'Cooling Tower Fan Electricity Energy')
pump_electric_outputs = 'Pump Electricity Energy'
people_gain_outputs = LoadBalance.PEOPLE_GAIN
solar_gain_outputs = LoadBalance.SOLAR_GAIN
infil_gain_outputs = LoadBalance.INFIL_GAIN
infil_loss_outputs = LoadBalance.INFIL_LOSS
vent_loss_outputs = LoadBalance.VENT_LOSS
vent_gain_outputs = LoadBalance.VENT_GAIN
nat_vent_gain_outputs = LoadBalance.NAT_VENT_GAIN
nat_vent_loss_outputs = LoadBalance.NAT_VENT_LOSS
all_output = \
[cooling_outputs, heating_outputs, lighting_outputs, electric_equip_outputs, gas_equip_outputs,
 process_outputs, shw_outputs, fan_electric_outputs, pump_electric_outputs,
 people_gain_outputs, solar_gain_outputs, infil_gain_outputs, infil_loss_outputs,
 vent_loss_outputs, vent_gain_outputs, nat_vent_gain_outputs, nat_vent_loss_outputs]


if all_required_inputs(ghenv.Component):
    # check the size of the SQL file to see if we should use the CLI
    assert os.path.isfile(_sql), 'No sql file found at: {}.'.format(_sql)
    if os.name == 'nt' and os.path.getsize(_sql) < 1e8:
        # small file on windows; use IronPython like usual
        # create the SQL result parsing object
        sql_obj = SQLiteResult(_sql)

        # get all of the results relevant for energy use
        cooling = sql_obj.data_collections_by_output_name(cooling_outputs)
        heating = sql_obj.data_collections_by_output_name(heating_outputs)
        lighting = sql_obj.data_collections_by_output_name(lighting_outputs)
        electric_equip = sql_obj.data_collections_by_output_name(electric_equip_outputs)
        hot_water = sql_obj.data_collections_by_output_name(shw_outputs)
        gas_equip = sql_obj.data_collections_by_output_name(gas_equip_outputs)
        process = sql_obj.data_collections_by_output_name(process_outputs)
        fan_electric = sql_obj.data_collections_by_output_name(fan_electric_outputs)
        pump_electric = sql_obj.data_collections_by_output_name(pump_electric_outputs)

        # get all of the results relevant for gains and losses
        people_gain = sql_obj.data_collections_by_output_name(people_gain_outputs)
        solar_gain = sql_obj.data_collections_by_output_name(solar_gain_outputs)
        infil_gain = sql_obj.data_collections_by_output_name(infil_gain_outputs)
        infil_loss = sql_obj.data_collections_by_output_name(infil_loss_outputs)
        vent_loss = sql_obj.data_collections_by_output_name(vent_loss_outputs)
        vent_gain = sql_obj.data_collections_by_output_name(vent_gain_outputs)
        nat_vent_gain = sql_obj.data_collections_by_output_name(nat_vent_gain_outputs)
        nat_vent_loss = sql_obj.data_collections_by_output_name(nat_vent_loss_outputs)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', _sql]
        for outp in all_output:
            out_str = json.dumps(outp) if isinstance(outp, tuple) else '["{}"]'.format(outp)
            cmds.append(out_str)
        use_shell = True if os.name == 'nt' else False
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
        stdout = process.communicate()
        data_coll_dicts = json.loads(stdout[0])

        # get all of the results relevant for energy use
        cooling = serialize_data(data_coll_dicts[0])
        heating = serialize_data(data_coll_dicts[1])
        lighting = serialize_data(data_coll_dicts[2])
        electric_equip = serialize_data(data_coll_dicts[3])
        gas_equip = serialize_data(data_coll_dicts[4])
        process = serialize_data(data_coll_dicts[5])
        hot_water = serialize_data(data_coll_dicts[6])
        fan_electric = serialize_data(data_coll_dicts[7])
        pump_electric = serialize_data(data_coll_dicts[8])

        # get all of the results relevant for gains and losses
        people_gain = serialize_data(data_coll_dicts[9])
        solar_gain = serialize_data(data_coll_dicts[10])
        infil_gain = serialize_data(data_coll_dicts[11])
        infil_loss = serialize_data(data_coll_dicts[12])
        vent_loss = serialize_data(data_coll_dicts[13])
        vent_gain = serialize_data(data_coll_dicts[14])
        nat_vent_gain = serialize_data(data_coll_dicts[15])
        nat_vent_loss = serialize_data(data_coll_dicts[16])

    # do arithmetic with any of the gain/loss data collections
    if len(infil_gain) == len(infil_loss):
        infiltration_load = subtract_loss_from_gain(infil_gain, infil_loss)
    if len(vent_gain) == len(vent_loss) == len(cooling) == len(heating):
        mech_vent_loss = subtract_loss_from_gain(heating, vent_loss)
        mech_vent_gain = subtract_loss_from_gain(cooling, vent_gain)
        mech_vent_load = [data.duplicate() for data in
                          subtract_loss_from_gain(mech_vent_gain, mech_vent_loss)]
        for load in mech_vent_load:
            load.header.metadata['type'] = \
                'Zone Ideal Loads Ventilation Heat Energy'
    if len(nat_vent_gain) == len(nat_vent_loss):
        nat_vent_load = subtract_loss_from_gain(nat_vent_gain, nat_vent_loss)

    # remove the district hot water system used for service hot water from space heating
    shw_equip, distr_i = [], None
    for i, heat in enumerate(heating):
        if not isinstance(heat, float):
            try:
                heat_equip = heat.header.metadata['System']
                if heat_equip.startswith('SHW'):
                    shw_equip.append(i)
                elif heat_equip == 'SERVICE HOT WATER DISTRICT HEAT':
                    distr_i = i
            except KeyError:
                pass
    if len(shw_equip) != 0 and distr_i is None:
        hot_water = [heating.pop(i) for i in reversed(shw_equip)]
    elif distr_i is not None:
        for i in reversed(shw_equip + [distr_i]):
            heating.pop(i)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Assign internal thermal masses to Rooms, which can be used to account for the
effects of furniture inside Rooms or other massive building components like
staircases, hearths, etc.
_
The component accepts either Rhino geometry (representing furniture or massive
elements) or a numerical value of the mass's surface area. Several of these
components can be used in a series to descibe different internal masses made
of different materials.
_
Note that internal masses assigned this way cannot "see" solar radiation that
may potentially hit them and, as such, caution should be taken when using this
component with internal mass objects that are not always in shade. Masses are
factored into the the thermal calculations of the Room by undergoing heat
transfer with the indoor air.
-

    Args:
        _rooms: Honeybee Rooms to which internal masses should be assigned.
        _geo_or_area: A list of Rhino breps or meshes representing the surfaces of internal
            masses that are exposed to the air of the Room. Alternatively, this can
            be a number or list of numbers representing the surface area of the
            internal masses (in square meters) that are exposed to the Room air.
            _
            In the case of Rhino geometry representing the surfaces of internal
            masses, this component will determine which Room the geometry is in.
            However, geometry must lie COMPLETELY inside a single Room and
            cannot span between Rooms or span outside the building. If a geometry
            lies between two Rooms, it should be split in two along the boundary
            between the Rooms. Also note that geometries are assumed to have only
            one side exposed to the Room air so, if they are meant to represent
            a 2-sided object, the geometry should be duplicated and offset.
            _
            In the case of numbers representing the the surface area of the
            internal masses, inputs can be either a single number (which will be
            used to put internal masses into all Rooms using the specified
            surface area), or it can be a list of numbers that matches the input
            Rooms, which can be used to assign different amounts of mass surface
            area to different Rooms. All numbers are assumed to be in square meters.
        _construction: An OpaqueConstruction object that represents the material
            that the internal thermal mass is composed of. This can also be
            text for the identifier of the construction within the library.
        _name_: An optional text name for the internal mass. This can be useful for
            keeping track of different internal masses when using several of
            these components in series. If unspecified, a unique one will be
            generated.

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with internal masses assigned.
"""

ghenv.Component.Name = 'HB Internal Mass'
ghenv.Component.NickName = 'InternalMass'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import honeybee_energy dependencies
    from honeybee_energy.internalmass import InternalMass
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_face3d
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list, \
        document_counter
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects and process the construction
    rooms = [room.duplicate() for room in _rooms]
    if isinstance(_construction, str):
        _construction = opaque_construction_by_identifier(_construction)

    # determine whether the input _geo_or_area is geometry or floats
    try:
        areas = [float(num) for num in _geo_or_area]
    except AttributeError:  # assume that the input is a list of geometry
        geo = [f for geo in _geo_or_area for f in to_face3d(geo)]
        conversion = conversion_to_meters() ** 2
        areas = [0 for room in rooms]
        for i, room in enumerate(rooms):
            for face in geo:
                if room.geometry.is_point_inside(face.center):
                    areas[i] += face.area * conversion

    # create the internal mass objects and assign them to the rooms
    for i, room in enumerate(rooms):
        area = longest_list(areas, i)
        if area != 0:
            if len(_name_) == 0:  # make a default Room name
                display_name = 'Internal Mass {}'.format(document_counter('mass_count'))
            else:
                display_name = '{}_{}'.format(longest_list(_name_, i), i + 1) \
                    if len(_name_) != len(_rooms) else longest_list(_name_, i)
            name = clean_ep_string(display_name)
            mass = InternalMass(name, _construction, area)
            mass.display_name = display_name
            room.properties.energy.add_internal_mass(mass)
            print('Internal mass with area {} m2 has been added to room '
                  '"{}"'.format(round(area, 3), room.display_name))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply ConstructionSet to Honeybee Rooms.
-

    Args:
        _rooms: Honeybee Rooms to which the input _constr_set should be assigned.
        _constr_set: A Honeybee ConstructionSet to be applied to the input _room.
            This can also be text for a construction set to be looked up in the
            construction set library.
    
    Returns:
        rooms: The input Rooms with their construction sets edited.
"""

ghenv.Component.Name = "HB Apply ConstructionSet"
ghenv.Component.NickName = 'ApplyConstrSet'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '3'


try:  # import the honeybee-energy extension
    from honeybee_energy.lib.constructionsets import construction_set_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    rooms = [obj.duplicate() for obj in _rooms]
    
    # process the input construction set if it's a string
    if isinstance(_constr_set, str):
        _constr_set = construction_set_by_identifier(_constr_set)
    
    # assign the construction set
    for rm in rooms:
        rm.properties.energy.construction_set = _constr_set



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct a complete thermal load balance from energy simulation results and
honeybee Rooms or a Model.

-
    Args:
        _rooms_model: An array of honeybee Rooms or a honeybee Model for which
            the thermal load balance will be computed. In most cases, these
            should be the Rooms or Model that are fed directly into the
            simulation. But this can also be a subset of such Rooms and the
            balance will only be computed for those Rooms.
        cooling_: Array of data collections for 'Zone Ideal Loads Supply Air
            Cooling Energy'.
        heating_: Array of data collections for 'Zone Ideal Loads Supply Air
            Heating Energy'.
        lighting_: Array of data collections for 'Zone Lights Heating Energy'.
        electric_equip_: Array of data collections for 'Zone Electric Equipment
            Heating Energy'.
        gas_equip_: Array of data collections for 'Zone Gas Equipment Heating Energy'.
        process_: Array of data collections for 'Zone Other Equipment Total
            Heating Energy'.
        hot_water_: Array of data collections for 'Water Use Equipment Zone Heat
            Gain Energy' that correspond to the input rooms.
        people_gain_: Array of data collections for 'Zone People Heating Energy'.
        solar_gain_: Array of data collections for 'Zone Windows Transmitted
            Solar Radiation Energy'.
        infiltration_load_: An array of data collections for the infiltration heat
            loss (negative) or heat gain (positive).
        mech_vent_load_: An array of data collections for the ventilation heat loss
            (negative) or heat gain (positive) as a result of meeting minimum
            outdoor air requirements with the mechanical system.
        nat_vent_load_: An array of data collections for the natural ventilation
            heat loss (negative) or heat gain (positive).
        face_energy_flow_: An array of data collections for the surface heat loss
            (negative) or heat gain (positive).

    Returns:
        report: ...
        balance:  A list of data collections where each collection represents a
            load balance term. This can then be plugged into the "LB Hourly Plot"
            or "LB Monthly Chart" to give a visualization of the load balance
            over all connected Rooms.
        balance_stor:  The balance output plus an additional term to represent
            the remainder of the load balance. This term is labeled "Storage" since
            it typically represents the energy being stored in the building's mass. 
            If this term is particularly large, it can indicate that not all of
            the load balance terms have been plugged into this component.
        norm_bal: A list of data collections where each collection represents a
            load balance term that has bee normalized by the Room floor area.
            This can then be plugged into the "LB Hourly Plot" or "LB Monthly Chart"
            to give a visualization of the load balance over all connected Rooms.
        norm_bal_stor:  The norm_bal output plus an additional term to represent
            the remainder of the load balance. This term is labeled "Storage" since
            it typically represents the energy being stored in the building's mass. 
            If this term is particularly large, it can indicate that not all of
            the load balance terms have been plugged into this component.
"""

ghenv.Component.Name = 'HB Thermal Load Balance'
ghenv.Component.NickName = 'LoadBalance'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.loadbalance import LoadBalance
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import units_system
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def check_input(input_list):
    """Check that an input isn't a zero-length list or None."""
    return None if len(input_list) == 0 or input_list[0] is None else input_list


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models
    is_model, floor_area = False, 0
    rooms = []
    for hb_obj in _rooms_model:
        if isinstance(hb_obj, Model):
            rooms.extend(hb_obj.rooms)
            is_model = True
            floor_area += hb_obj.floor_area
        else:
            rooms.append(hb_obj)

    # if the input is for individual rooms, check the solar to ensure no groued zones
    if not is_model and len(solar_gain_) != 0:
        msg = 'Air boundaries with grouped zones detected in solar data but individual ' \
            'rooms were input.\nIt is recommended that the full model be input for ' \
            '_rooms_model to ensure correct representaiton of solar.'
        for coll in solar_gain_:
            if 'Solar Enclosure' in coll.header.metadata['Zone']:
                print(msg)
                give_warning(ghenv.Component, msg)

    # process all of the inputs
    cooling_ = check_input(cooling_)
    heating_ = check_input(heating_)
    lighting_ = check_input(lighting_)
    electric_equip_ = check_input(electric_equip_)
    gas_equip_ = check_input(gas_equip_)
    hot_water_ = check_input(hot_water_)
    people_gain_ = check_input(people_gain_)
    solar_gain_ = check_input(solar_gain_)
    infiltration_load_ = check_input(infiltration_load_)
    mech_vent_load_ = check_input(mech_vent_load_)
    nat_vent_load_ = check_input(nat_vent_load_)
    face_energy_flow_ = check_input(face_energy_flow_)

    # process hot water to ensure it's the correct type
    hw_type = 'Water Use Equipment Heating Energy'
    if hot_water_ is not None and hot_water_[0].header.metadata['type'] == hw_type:
        hot_water_ = [hw.duplicate() * 0.25 for hw in hot_water_]
        for hw in hot_water_:
            hw.header.metadata = {
                'type': 'Water Use Equipment Zone Sensible Heat Gain Energy',
                'System': hw.header.metadata['System']
            }

    # construct the load balance object and output the results
    load_bal_obj = LoadBalance(
        rooms, cooling_, heating_, lighting_, electric_equip_, gas_equip_, process_,
        hot_water_, people_gain_, solar_gain_, infiltration_load_, mech_vent_load_,
        nat_vent_load_, face_energy_flow_, units_system(), use_all_solar=is_model)
    if is_model:
        load_bal_obj.floor_area = floor_area

    balance = load_bal_obj.load_balance_terms(False, False)
    if len(balance) != 0:
        balance_stor = balance + [load_bal_obj.storage]
        norm_bal = load_bal_obj.load_balance_terms(True, False)
        norm_bal_stor = load_bal_obj.load_balance_terms(True, True)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Compute spatially-resolved operative temperature and Predicted Mean Vote (PMV)
thermal comfort from a Honeybee model. This recipe can also (optionally)
compute Standard Effective Temperature (SET).
_
This recipe uses EnergyPlus to obtain surface temperatures and indoor air
temperatures + humidities. Outdoor air temperatures, relative humidities, and
air speeds are taken directly from the EPW. The energy properties of the model
geometry are what determine the outcome of the simulation, though the model's
Radiance sensor grids are what determine where the comfort mapping occurs.
_
Longwave radiant temperatures are obtained by computing spherical view factors
from each sensor to the Room surfaces of the model using Radiance. These view factors
are then multiplied by the surface temperatures output by EnergyPlus to yield
longwave MRT at each sensor. All indoor shades (eg. those representing furniture)
are assumed to be at the room-average MRT.
_
A Radiance-based enhanced 2-phase method is used for all shortwave MRT calculations,
which precisely represents direct sun by tracing a ray from each sensor to the
solar position. To determine Thermal Comfort Percent (TCP), the occupancy schedules
of the energy model are used. Any hour of the occupancy schedule that is 0.1 or
greater will be considered occupied. All hours of the outdoors are considered occupied.
-

    Args:
        _model: A Honeybee Model for which PMV comfort will be mapped. Note that
            this model should have radiance grids assigned to it in order
            to produce meaningful results.
        _epw: Path to an EPW weather file to be used for the comfort map simulation.
        _ddy: Path to a DDY file with design days to be used for the initial sizing
            calculation of the energy simulation.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        run_period_: An AnalysisPeriod to set the start and end dates of the simulation.
            If None, the simulation will be annual.
        add_str_: THIS OPTION IS FOR ADVANCED USERS OF ENERGYPLUS. You can
            input additional text strings here to be appended to the IDF before
            energy simulation.  The input should be complete EnergyPlus objects
            following the IDF format. This input can be used to write objects
            into the IDF that are not currently supported by Honeybee.
        write_set_map_: A boolean to note whether the output temperature CSV should
            record Operative Temperature or Standard Effective Temperature (SET).
            SET is relatively intense to compute and so only recording Operative
            Temperature can greatly reduce run time, particularly when air speeds
            are low. However, SET accounts for all 6 PMV model inputs and so is a
            more representative "feels-like" temperature for the PMV model.
        _air_speed_: A single number for air speed in m/s or an hourly data collection
            of air speeds that align with the input run_period_. This will be
            used for all indoor comfort evaluation. Note that the EPW wind speed
            will be used for any outdoor sensors. (Default: 0.1).
        _met_rate_: A single number for metabolic rate in met or an hourly data collection
            of met rates that align with the run_period_. (Default: 1.1, for
            seated, typing).
        _clo_value_: A single number for clothing level in clo or an hourly data collection
            of clothing levels that align with the run_period_. (Default: 0.7,
            for pants and a long sleeve shirt).
        comfort_par_: Optional comfort parameters from the "LB PMV Comfort Parameters"
            component to specify the criteria under which conditions are
            considered acceptable/comfortable. The default will assume a
            PPD threshold of 10% and no absolute humidity constraints.
        solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        env_conds: A folder containing CSV matrices with all of the environmental conditions
            that were input to the comfort model. These can be loaded into Grasshopper
            using the "HB Read Environment Matrix" component. This includes the following.
                * MRT
                * Air Temperature
                * Relative Humidity
                * Longwave MRT
                * Shortwave MRT Delta
        temperature: A folder containing CSV maps of Operative Temperature for each sensor
            grid at each time step of the analysis. Alternatively, if the
            write_set_map_ option is used, the CSV maps here will contain
            Standard Effective Temperature (SET). This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are in Celsius.
        condition: A folder containing CSV maps of comfort conditions for each sensor
            grid at each time step of the analysis. This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are as follows.
                * -1 = unacceptably cold conditions
                *  0 = neutral (comfortable) conditions
                * +1 = unacceptably hot conditions
        pmv: A folder containing CSV maps of the Predicted Mean Vote (PMV) for each
            sensor grid at each time step of the analysis. This can be connected
            to the "HB Read Thermal Matrix" component to parse detailed results
            into Grasshopper. This can be used to understand not just whether
            conditions are acceptable but how uncomfortably hot or cold they are.
        TCP: Lists of values between 0 and 100 for the Thermal Comfort Percent (TCP).
            These can be plugged into the "LB Spatial Heatmap" component along
            with meshes of the sensor grids to visualize spatial thermal comfort.
            TCP is the percentage of occupied time where thermal conditions are
            acceptable/comfortable. Occupied hours are determined from the
            occupancy schedules of each room (any time where the occupancy
            schedule is >= 0.1 will be considered occupied). Outdoor sensors
            are considered occupied at all times. More custom TCP studies can
            be done by post-processing the condition results.
        HSP: Lists of values between 0 and 100 for the Heat Sensation Percent (HSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably hot locations.
            HSP is the percentage of occupied time where thermal conditions are
            hotter than what is considered acceptable/comfortable. Occupied hours
            are determined from the occupancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom HSP
            studies can be done by post-processing the condition results.
        CSP: Lists of values between 0 and 100 for the Cold Sensation Percent (CSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably cold locations.
            CSP is the percentage of occupied time where thermal conditions are
            colder than what is considered acceptable/comfortable. Occupied hours
            are determined from the occupancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom CSP
            studies can be done by post-processing the condition results.
"""

ghenv.Component.Name = 'HB PMV Comfort Map'
ghenv.Component.NickName = 'PMVMap'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import units_system
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('pmv_comfort_map')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('epw', _epw)
    recipe.input_value_by_name('ddy', _ddy)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('run-period', run_period_)
    recipe.input_value_by_name('additional-idf', add_str_)
    recipe.input_value_by_name('write-set-map', write_set_map_)
    recipe.input_value_by_name('air-speed', _air_speed_)
    recipe.input_value_by_name('met-rate', _met_rate_)
    recipe.input_value_by_name('clo-value', _clo_value_)
    recipe.input_value_by_name('comfort-parameters', comfort_par_)
    recipe.input_value_by_name('solarcal-parameters', solar_body_par_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # perform an extra check for units because many people forget to check them
    if isinstance(_model, Model):
        check_model = _model
        if check_model.units != 'Meters':
            check_model = _model.duplicate()
            check_model.convert_to_units('Meters')
        # remove degenerate geometry within native E+ tolerance of 0.01 meters
        for room in check_model.rooms:
            try:
                room.remove_colinear_vertices_envelope(
                    tolerance=0.01, delete_degenerate=True)
            except AssertionError as e:  # room removed; likely wrong units
                error = 'Your Model units system is: {}. ' \
                    'Is this correct?\n{}'.format(_model.units, e)
                raise ValueError(error)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(
        run_settings_, radiance_check=True, openstudio_check=True, silent=silent)

    # load the results
    try:
        env_conds = recipe_result(recipe.output_value_by_name('environmental-conditions', project_folder))
        temperature = recipe_result(recipe.output_value_by_name('temperature', project_folder))
        condition = recipe_result(recipe.output_value_by_name('condition', project_folder))
        pmv = recipe_result(recipe.output_value_by_name('pmv', project_folder))
        TCP = recipe_result(recipe.output_value_by_name('tcp', project_folder))
        HSP = recipe_result(recipe.output_value_by_name('hsp', project_folder))
        CSP = recipe_result(recipe.output_value_by_name('csp', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get all the data within a table of a Summary Report using the table name.
_
All of the avaialable tables can be browsed by opening the .html output from the
simulation in a web browser.
-

    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.
        _table_name: Text string for the name of a table of a Summary Report.
            Examples include: General, Utility Use Per Conditioned Floor Area,
            and many more options that can be browsed in the .html file.

    Returns:
        values: A data tree represening the table matrix, with each branch (sub-list)
            of the tree representing a row of the table and each index of each
            branch corresponding to a value in a column. The order of outputs
            should reflect how the table appears in the HTML output. Note that
            any energy values in MJ or GJ in the .html output will automatically
            be converted to kWh on import.
        col_names: A list of text for the names of each of the columns in the table.
            These order of this list corresponds directly to the order of items
            each of the values sub-list
        row_names: A list of text for the names of each of the rows of the table.
            Each name in this list corresponds to a branch in the output values
            data tree.
"""

ghenv.Component.Name = 'HB Read Tabular Data'
ghenv.Component.NickName = 'ReadTable'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import os
import subprocess
import json

try:
    from ladybug.sql import SQLiteResult
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(_sql)  # create the SQL result parsing object
        results = sql_obj.tabular_data_by_name(_table_name)
        values = list_to_data_tree(list(results.values()))
        row_names = list(results.keys())
        col_names = sql_obj.tabular_column_names(_table_name)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'tabular-data', _sql, _table_name]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        results = json.loads(stdout[0])
        values = list_to_data_tree(results)
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'tabular-metadata', _sql, _table_name]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        metadata_dict = json.loads(stdout[0])
        row_names = metadata_dict['row_names']
        col_names = metadata_dict['column_names']



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply absolute load values to Rooms.
_
Note that, while the assigned load values are absolute, this component will convert
them to the "normalized" value for each room (eg. lighting per floor area) in
order to apply them to the rooms. So any existing specification of load intensity
is overwritten with the absolute load here.
_
This also means that, if a room has no floors (or exterior walls for infiltration),
the resulting load values will be equal to 0 regardless of the input here. The
only exception is the vent_flow_, which will be applied regardless of the room
properties.
_
This component will not edit any of the schedules or other properties associated
with each load value. If no schedule currently exists to describe how the load
varies over the simulation, the "Always On" schedule will be used as a default.
-

    Args:
        _rooms: Honeybee Rooms to which the input load values should be assigned.
        person_count_: A number for the quantity of people in the room.
        lighting_watts_: A number for the installed wattage of lighting in the room (W).
        electric_watts_: A number for the installed wattage of electric equipment
            in the room (W).
        gas_watts_: A number for the installed wattage of gas equipment in the room (W).
        hot_wtr_flow_: Number for the peak flow rate of service hot water in the
            room in liters per hour (L/h).
        infiltration_ach_: A number for the infiltration flow rate in air changes
            per hour (ACH).
        vent_flow_: A numerical value for the absolute of flow of outdoor air ventilation
            for the room in cubic meters per second (m3/s). Note that inputting
            a value here will overwrite all specification of outdoor air ventilation
            currently on the room (per_floor, per_person, ach).

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with their load values modified.
"""

ghenv.Component.Name = 'HB Apply Abolute Load Values'
ghenv.Component.NickName = 'AbsoluteLoadVals'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
    from ladybug_rhino.config import conversion_to_meters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    conversion = conversion_to_meters()
    rooms = [room.duplicate() for room in _rooms]  # duplicate the initial objects

    # assign the person_count_
    if len(person_count_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_people(
                longest_list(person_count_, i), conversion)

    # assign the lighting_watts_
    if len(lighting_watts_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_lighting(
                longest_list(lighting_watts_, i), conversion)

    # assign the electric_watts_
    if len(electric_watts_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_electric_equipment(
                longest_list(electric_watts_, i), conversion)

    # assign the gas_watts_
    if len(gas_watts_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_gas_equipment(
                longest_list(gas_watts_, i), conversion)

    # assign the hot_wtr_flow_
    if len(hot_wtr_flow_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_service_hot_water(
                longest_list(hot_wtr_flow_, i), conversion)

    # assign the infiltration_ach_
    if len(infiltration_ach_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_infiltration_ach(
                longest_list(infiltration_ach_, i), conversion)

    # assign the vent_flow_
    if len(vent_flow_) != 0:
        for i, room in enumerate(rooms):
            room.properties.energy.absolute_ventilation(longest_list(vent_flow_, i))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Lighting object into its constituient properties.
-

    Args:
        _lighting: A Lighting object to be deconstructed.
    
    Returns:
        name: Text string for the lighting display name.
         watts_per_area: A numerical value for the lighting power density in
            Watts per square meter of floor area.
        schedule: A fractional for the use of lights over the course of the year.
            The fractional values will get multiplied by the watts_per_area to
            yield a complete lighting profile.
        radiant_fract: A number between 0 and 1 for the fraction of the total
            lighting load given off as long wave radiant heat.
        visible_fract: A number between 0 and 1 for the fraction of the total
            lighting load given off as short wave visible light.
        return_fract: A number between 0 and 1 for the fraction of the total
            lighting load that goes into the zone return air.
        baseline: The baseline lighting power density in W/m2 of floor area. This
            baseline is useful to track how much better the installed lights
            are in comparison to a standard like ASHRAE 90.1.
"""

ghenv.Component.Name = "HB Deconstruct Lighting"
ghenv.Component.NickName = 'DecnstrLighting'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.lighting import Lighting
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _lighting is not None:
    # check the input
    assert isinstance(_lighting, Lighting), \
        'Expected Lighting object. Got {}.'.format(type(_lighting))

    # get the properties of the object
    name = _lighting.display_name
    watts_per_area = _lighting.watts_per_area
    schedule = _lighting.schedule
    radiant_fract = _lighting.radiant_fraction
    visible_fract = _lighting.visible_fraction
    return_fract = _lighting.return_air_fraction
    baseline = _lighting.baseline_watts_per_area



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a schedule defined by a list of values at a fixed interval or timestep
running over the entirety of the simulation period.
-

    Args:
        _values: A list of timeseries values occuring at a fixed timestep over
            the entire simulation. Typically, this should be a list of 8760
            values for each hour of the year but it can be a shorter list if
            you don't plan on using the schedule in an annual simulation. In
            this case, the analysis_period_ should probably be different than
            the default. This list can also have a length much greater than
            8760 if a timestep greater than 1 is used.
        _timestep_: An integer for the number of steps per hour that the input
            values correspond to.  For example, if each value represents 30
            minutes, the timestep is 2. For 15 minutes, it is 4. Default is 1,
            meaning each value represents a single hour. Must be one of the
            following: (1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60).
        analysis_period_: A ladybug AnalysisPeriod object to note when the input
            values take effect. Default is for the whole year. Note that this
            default usually should not be changed unless you plan to run a
            simulation that is much shorter than a year.
        _name_: Text to set the name for the Schedule and to be incorporated
            into a unique Schedule identifier.
        _type_limit_: A text string from the name of the ScheduleTypeLimit to
            be looked up in the schedule type limit library. This can also be a
            custom ScheduleTypeLimit object from the "HB Type Limit" component.
            The input here will be used to validate schedule values against
            upper/lower limits and assign units to the schedule values. Default:
            "Fractional" for values that range continuously between 0 and 1.
            Choose from the following built-in options:
                * Fractional
                * On-Off
                * Temperature
                * Activity Level
                * Power
                * Humidity
                * Angle
                * Delta Temperature
                * Control Level

    Returns:
        report: Reports, errors, warnings, etc.
        schedule: A ScheduleRuleset object that can be assigned to a Room, a Load
            object, or a ProgramType object.
"""

ghenv.Component.Name = 'HB Fixed Interval Schedule'
ghenv.Component.NickName = 'FixedIntervalSchedule'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import the ladybug dependencies
    from ladybug.dt import Date
    from ladybug.analysisperiod import AnalysisPeriod
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.fixedinterval import ScheduleFixedInterval
    from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the defaults
    _timestep_ = 1 if _timestep_ is None else _timestep_
    start_date = Date(1, 1) if analysis_period_ is None else \
        analysis_period_.st_time.date
    name = clean_and_id_ep_string('FixedIntervalSchedule') if _name_ is None else \
        clean_ep_string(_name_)

    # perform a check to see if the input values align with the analysis period
    a_per = AnalysisPeriod() if analysis_period_ is None else analysis_period_
    if len(a_per) * _timestep_ != len(_values):
        msg = 'The number of values implied by the analysis period [{}]\n' \
            'is not the same as the number of values supplied [{}].\n' \
            'This may result in unexpeted behavior.'.format(
                len(a_per) * _timestep_, len(_values))
        print(msg)
        give_warning(ghenv.Component, msg)

    # get the ScheduleTypeLimit object
    if _type_limit_ is None:
        _type_limit_ = schedule_type_limit_by_identifier('Fractional')
    elif isinstance(_type_limit_, str):
        _type_limit_ = schedule_type_limit_by_identifier(_type_limit_)

    # create the schedule object
    schedule = ScheduleFixedInterval(
        name, _values, _type_limit_, _timestep_, start_date,
        placeholder_value=0, interpolate=False)
    if _name_ is not None:
        schedule.display_name = _name_



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse a zone sizing (ZSZ) csv result file from an energy simulation to get data
collections for the cooling/heating load over the peak design day.
-

    Args:
        _zsz: Full path to a zone sizing (ZSZ) csv result file that was
            generated by EnergyPlus.
    
    Returns:
        cooling_load: a list of HourlyContinuousCollections for zone cooling load.
            There will be one data collection per conditioned zone in the model.
        heating_load: a list of HourlyContinuousCollections for zone heating load.
            There will be one data collection per conditioned zone in the model.
"""

ghenv.Component.Name = 'HB Read Zone Sizing'
ghenv.Component.NickName = 'ReadZSZ'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from honeybee_energy.result.zsz import ZSZ
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    zsz_obj = ZSZ(_zsz)
    cooling_load = zsz_obj.cooling_load_data
    heating_load = zsz_obj.heating_load_data



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create an EnergyPlus window construction that includes shades/blinds or a dynamically-
controlled glass pane.
_
The result can be assigned to any Aperture or ConstructionSet just like any other
WindowConstruction.
-

    Args:
        _name_: Text to set the name for the Construction and to be incorporated
            into a unique Construction identifier.
        _win_constr: A WindowConstruction object that serves as the "switched off"
            version of the construction (aka. the "bare construction"). The
            shade material and shade location will be used to modify this
            starting construction. This can also be text for a construction
            identifier to be looked up in the window construction library.
        _shd_material: An Shade Material or an Blind Material that serves as the
            shading layer for this construction. This can also be a Glass Material,
            which will indicate that the WindowConstruction has a dynamically-
            controlled glass pane like an electrochromic window assembly. This
            can also be text for a material identifier to be looked up in the
            window material library.
        _shd_location_: Text to indicate where in the window assembly the shade material
            is located. (Default: "Interior"). Choose from the following 3 options:
            * Interior
            * Between
            * Exterior
            Note that the WindowConstruction must have at least one gas gap to use
            the "Between" option. Also note that, for a WindowConstruction with more
            than one gas gap, the "Between" option defalts to using the inner gap
            as this is the only option that EnergyPlus supports.
        _control_type_: An integer or text to indicate how the shading device is controlled,
            which determines when the shading is “on” or “off.” (Default: "AlwaysOn").
             Choose from the options below (units for the values of the corresponding
             setpoint are noted in parentheses next to each option):
             0 = AlwaysOn
             1 = OnIfHighSolarOnWindow (W/m2)
             2 = OnIfHighHorizontalSolar (W/m2)
             3 = OnIfHighOutdoorAirTemperature (C)
             4 = OnIfHighZoneAirTemperature (C)
             5 = OnIfHighZoneCooling (W)
             6 = OnNightIfLowOutdoorTempAndOffDay (C)
             7 = OnNightIfLowInsideTempAndOffDay (C)
             8 = OnNightIfHeatingAndOffDay (W)
        setpoint_: A number that corresponds to the specified control_type. This can
            be a value in (W/m2), (C) or (W) depending upon the control type.
        schedule_: An optional ScheduleRuleset or ScheduleFixedInterval to be applied
            on top of the control type. If None, the control type will govern all
            behavior of the construction.
    
    Returns:
        constr: A shaded window construction that can be assigned to Honeybee
            Apertures or ConstructionSets.
"""

ghenv.Component.Name = 'HB Window Construction Shade'
ghenv.Component.NickName = 'WindowConstrShd'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.construction.windowshade import WindowConstructionShade
    from honeybee_energy.lib.constructions import window_construction_by_identifier
    from honeybee_energy.lib.materials import window_material_by_identifier
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# dictionary to convert to acceptable control types
CONTROL_TYPES = {
    '0': 'AlwaysOn',
    '1': 'OnIfHighSolarOnWindow',
    '2': 'OnIfHighHorizontalSolar',
    '3': 'OnIfHighOutdoorAirTemperature',
    '4': 'OnIfHighZoneAirTemperature',
    '5': 'OnIfHighZoneCooling',
    '6': 'OnNightIfLowOutdoorTempAndOffDay',
    '7': 'OnNightIfLowInsideTempAndOffDay',
    '8': 'OnNightIfHeatingAndOffDay',
    'AlwaysOn': 'AlwaysOn',
    'OnIfHighSolarOnWindow': 'OnIfHighSolarOnWindow',
    'OnIfHighHorizontalSolar': 'OnIfHighHorizontalSolar',
    'OnIfHighOutdoorAirTemperature': 'OnIfHighOutdoorAirTemperature',
    'OnIfHighZoneAirTemperature': 'OnIfHighZoneAirTemperature',
    'OnIfHighZoneCooling': 'OnIfHighZoneCooling',
    'OnNightIfLowOutdoorTempAndOffDay': 'OnNightIfLowOutdoorTempAndOffDay',
    'OnNightIfLowInsideTempAndOffDay': 'OnNightIfLowInsideTempAndOffDay',
    'OnNightIfHeatingAndOffDay': 'OnNightIfHeatingAndOffDay'
    }


if all_required_inputs(ghenv.Component):
    # set default values
    constr_id = clean_and_id_ep_string('ShadedWindowConstruction') if _name_ is None else \
        clean_ep_string(_name_)
    _shd_location_ = 'Interior' if _shd_location_ is None else _shd_location_.title()
    _control_type_ = 'AlwaysOn' if _control_type_ is None \
        else CONTROL_TYPES[_control_type_]

    # get objects from the library if they are strings
    if isinstance(_win_constr, str):
        win_con = window_construction_by_identifier(_win_constr)
        # duplicate and rename to avoid having the same construction name in one model
        _win_constr = win_con.duplicate()
        _win_constr.identifier = '{}_Unshaded'.format(constr_id)
    if isinstance(_shd_material, str):
        _shd_material = window_material_by_identifier(_shd_material)
    if isinstance(schedule_, str):
        schedule_ = schedule_by_identifier(schedule_)

    # create the construction object
    constr = WindowConstructionShade(
        constr_id, _win_constr, _shd_material, _shd_location_, _control_type_,
        setpoint_, schedule_)
    if _name_ is not None:
        constr.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Get a ladybug DataCollection representing this schedule at a given timestep.
_
This DataCollection can be used to visualize timeseries schedule values over
the entire period of a simulation using a component like the "LB Hourly Plot".
_
This DataCollection can also be used in the crafting of conditional statements
with the ladybug components. For example, making a psychrometric chart of zone
temperature/humidity for only the hours that the occupancy schedule is above a
certain threshold.
-

    Args:
        _schedule: A ScheduleRuleSet or SchedileFixedInterval for which a DataCollection
            of timeseries schedule values will be produced. This can also be the
            identifier of a Schedule to be looked up in the schedule library.
        analysis_period_: An optional AnalysisPeriod to set the start and end datetimes
            of the resulting DataCollection. The timestep of this AnalysisPeriod
            will also be used to determine the timestep of the resulting DataCollection.
        _week_start_day_: An optional text string to set the start day of the week
            of the schedule timeseries data. Default: Sunday. Choose from the following:
                * Sunday
                * Monday
                * Tuesday
                * Wednesday
                * Thursday
                * Friday
                * Saturday
        holidays_: An optional list of strings (eq: 25 Dec) to represent the holidays
            in the resulting timeseries. Holiday schedules will be used for these
            dates in the resulting timeseries.
    
    Returns:
        data: A ladybug DataCollection representing the timeseries values of the schedule.
            This can be used to visualize timeseries schedule values over the entire
            period of a simulation using a component like the "LB Hourly Plot".
"""

ghenv.Component.Name = "HB Schedule to Data"
ghenv.Component.NickName = 'SchToData'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = "2"


try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the ladybug dependencies
    from ladybug.dt import Date, DateTime
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the schedue from the library if it's a string
    if isinstance(_schedule, str):
        _schedule = schedule_by_identifier(_schedule)

    # process the _week_start_day_
    week_start_day = 'Sunday' if _week_start_day_ is None else _week_start_day_.title()

    # process the analysis period if it is input
    if analysis_period_ is not None:
        start_date = analysis_period_.st_time.date
        end_date = analysis_period_.end_time.date
        timestep = analysis_period_.timestep
    else:
        start_date, end_date, timestep = Date(1, 1), Date(12, 31), 1

    # process the holidays_ if they are input
    holidays = None
    if len(holidays_) != 0 and holidays_[0] is not None:
        try:
            holidays = tuple(Date.from_date_string(hol) for hol in holidays_)
        except ValueError:
            holidays = tuple(DateTime.from_date_time_string(hol).date for hol in holidays_)

    # create the DataCollection
    if isinstance(_schedule, ScheduleRuleset):
        data = _schedule.data_collection(
            timestep, start_date, end_date, week_start_day, holidays, leap_year=False)
    else:  # assume that it is a ScheduleFixedInterval
        data = _schedule.data_collection_at_timestep(timestep, start_date, end_date)

    # if there are hour inputs on the analysis_period_, apply it to the data
    if analysis_period_ is not None and \
            (analysis_period_.st_hour != 0 or analysis_period_.end_hour != 23):
        data = data.filter_by_analysis_period(analysis_period_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Load OpenStudio measures into Grasshopper and assign the measure's input arguments.
The resulting measure object can be plugged into the "measures_" input of the
"HB Model To OSM" component in order to be included in the export to OpenStudio.
-
Read more about OpenStudio measures and creating your own here:
http://nrel.github.io/OpenStudio-user-documentation/reference/measure_writing_guide/
-
You can also download several measures created by others from here:
https://bcl.nrel.gov/nrel/types/measure
-


    Args:
        _measure_path: Path to the folder in which the measure exists. This folder
            must contain a measure.rb and a measure.xml file. Note that connecting
            an input here will transform the component, essentially removing this
            input and changing all of the other component inputs to be input
            arguments for the measure.
    
    Returns:
        measure: A measure measure object can be plugged into the "measures_" input
            of the "HB Model To OSM" component in order to be included in the
            export to OpenStudio.
"""

ghenv.Component.Name = 'HB Load Measure'
ghenv.Component.NickName = 'LoadMeasure'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from honeybee_energy.measure import Measure
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import Grasshopper.Kernel as gh


def add_component_input_from_arg(argument):
    """Add an input parameter to this component using a MeasureArgument object.

    Args:
        argument: A honeybee-energy MeasureArgument object that will be used
            to create a new input parameter for the component.
    """
    # create the input parameter object
    param = gh.Parameters.Param_ScriptVariable()

    # assign the required properties to the input
    arg_id = '_{}'.format(argument.identifier) if argument.required else \
        '{}_'.format(argument.identifier)
    param.NickName = arg_id
    param.Name = arg_id if argument.display_name is None else argument.display_name
    param.AllowTreeAccess = False  # Measures don't support lists
    param.Access = gh.GH_ParamAccess.item  # Measures don't support lists
    param.Optional = True  # always needed so that default can come from measure file

    # assign the optional properties to the input if they exist
    if argument.type_text == 'Choice':
        descr = [argument.description] if argument.description else []
        if None not in argument.valid_choices:
            descr.append('Choose from the following options:')
            descr = descr + list(argument.valid_choices)
        param.Description = '\n '.join(descr)
    elif argument.description:
        param.Description = argument.description
    if argument.default_value is not None:
        param.AddVolatileData(gh.Data.GH_Path(0), 0, argument.default_value)
    elif argument.type_text == 'Choice' and argument.valid_choices == (None,):
        param.AddVolatileData(gh.Data.GH_Path(0), 0, 'None')

    # add the parameter to the compoent
    index = ghenv.Component.Params.Input.Count
    ghenv.Component.Params.RegisterInputParam(param, index)
    ghenv.Component.Params.OnParametersChanged()

def transform_name_and_description(measure):
    """Transform this component's name and description to match a measure.

    Args:
        measure: A honeybee-energy Measure object that will be used to assign
            this component's name and description.
    """
    # assign the measure metadata
    ghenv.Component.NickName = measure.identifier
    ghenv.Component.Name = measure.display_name if measure.display_name \
        else measure.identifier
    if measure.description:
        ghenv.Component.Description = measure.description


def transform_component(measure):
    """Transform this component to have a name and arguments that match a measure.

    Args:
        measure: A honeybee-energy Measure object that will be used to assign
            this component's name, description, and input arguments.
    """
    # assign the measure metadata
    transform_name_and_description(measure)
    # assign the input arguments
    for arg in measure.arguments:
        add_component_input_from_arg(arg)


def check_arguments_and_set_defaults(measure):
    """Check to be sure the names of component inputs align with measure arguments.

    This method will also assign any default values from the measure if there is
    no value input to the component.

    Args:
        measure: A honeybee-energy Measure object that will be used check this
            component's input arguments.
    """
    for i in range(1, ghenv.Component.Params.Input.Count):
        # get the param and the measure argument object
        param = ghenv.Component.Params.Input[i]
        arg = measure_init.arguments[i - 1]

        # check that the param matches the measure argument
        assert arg.identifier in param.NickName, \
            "This component's inputs do not match that of the input measure.\n" \
            "Grab a fresh 'HB Load Measure' component and reload the measure."

        # add any default values
        if not param.VolatileDataCount or param.VolatileData[0][0] is None:
            if arg.default_value is not None:
                param.AddVolatileData(gh.Data.GH_Path(0), 0, arg.default_value)
            elif arg.type_text == 'Choice' and arg.valid_choices == (None,):
                param.AddVolatileData(gh.Data.GH_Path(0), 0, 'None')


def update_measure_arguments(measure):
    """Update the arguments of a measure object based on this component's inputs.

    Args:
        measure: A honeybee-energy Measure object to have its arguments updated
            with the inputs to this component.
    """
    for i in range(1, ghenv.Component.Params.Input.Count):
        try:
            try:  # sense if there are multiple inputs
                run_count = ghenv.Component.RunCount
                value = ghenv.Component.Params.Input[i].VolatileData[0][run_count - 1]
            except Exception:
                vdc = ghenv.Component.Params.Input[i].VolatileDataCount
                value = ghenv.Component.Params.Input[i].VolatileData[0][vdc -  1]
            if value is not None:
                val = str(value)  # cast to string to avoid weird Grasshopper types
                argument = measure.arguments[i - 1]
                if argument.valid_choices == (None,) and val == 'None':
                    pass  # choice argument with no valid choices
                else:
                    argument.value = val if val != 'False' else False
        except Exception:  # there is no input for this value; just ignore it
            pass


def is_measure_input():
    """Check if a measure path is input to this component.

    This is needed because we don't know if there are default values for all
    required inputs until we load the measure.
    """
    if _measure_path is None:
        msg = 'Input parameter _measure_path failed to collect data!'
        print(msg)
        give_warning(ghenv.Component, msg)
        return False
    return True


if is_measure_input():
    # load the measure
    measure_init = Measure(_measure_path)

    # transform the component or check the inputs and set defaults
    if ghenv.Component.Params.Input.Count == 1:  # first time loading the measure
        transform_component(measure_init)
    else:  # the component has already been transformed
        transform_name_and_description(measure_init)
        check_arguments_and_set_defaults(measure_init)

    # if the measure has all inputs that it needs, output the measure
    if all_required_inputs(ghenv.Component):
        update_measure_arguments(measure_init)
        measure = measure_init



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply or edit load values on a Room or ProgramType.
_
This component will not edit any of the schedule objects associated with each load
value. If no schedule currently exists to describe how the load varies over the
simulation, the "Always On" schedule will be used as a default.
-

    Args:
        _room_or_program: Honeybee Rooms or ProgramType objects to which the input
            load objects should be assigned. This can also be the identifier of a
            ProgramType to be looked up in the program type library.
        people_per_floor_: A numerical value for the number of people per square
            meter of floor area.
        lighting_per_floor_: A numerical value for the lighting power density in
            Watts per square meter of floor area.
        electric_per_floor_: A numerical value for the electric equipment power density
            in Watts per square meter of floor area.
        gas_per_floor_: A numerical value for the gas equipment power density in
            Watts per square meter of floor area.
        hot_wtr_per_floor_: A numerical value for the total volume flow rate of water
            per unit area of floor in (L/h-m2).
        infilt_per_exterior_: A numerical value for the intensity of infiltration
            in m3/s per square meter of exterior surface area. Typical values for
            this property are as follows (note all values are at typical building
            pressures of ~4 Pa):
                * 0.0001 (m3/s per m2 facade) - Tight building
                * 0.0003 (m3/s per m2 facade) - Average building
                * 0.0006 (m3/s per m2 facade) - Leaky building
        vent_per_floor_: A numerical value for the intensity of outdoor air ventilation
            in m3/s per square meter of floor area. This will be added to the
            vent_per_person_ and vent_ach_ to produce the final minimum outdoor
            air specification.
        vent_per_person_: A numerical value for the intensity of outdoor air ventilation
            in m3/s per person. This will be added to the vent_per_floor_,
            and vent_ach_ to produce the final minimum outdoor air
            specification. Note that setting this value does not mean
            that ventilation is varied based on real-time occupancy but rather
            that the minimum level of ventilation is determined using this value
            and the People object of the zone. To vary ventilation on a timestep
            basis, a ventilation schedule should be used or the dcv_ option
            should be selected on the HVAC system if it is available.
        vent_ach_: A numerical value for the intensity of outdoor air ventilation in air
            changes er hour (ACH). This will be added to the vent_per_floor_
            and vent_per_person_ to produce the final minimum outdoor air
            specification.

    Returns:
        report: Reports, errors, warnings, etc.
        mod_obj: The input Rooms or ProgramTypes with their load values modified.
"""

ghenv.Component.Name = "HB Apply Load Values"
ghenv.Component.NickName = 'ApplyLoadVals'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "2"

import uuid

try:
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.people import People
    from honeybee_energy.load.lighting import Lighting
    from honeybee_energy.load.equipment import ElectricEquipment, GasEquipment
    from honeybee_energy.load.hotwater import ServiceHotWater
    from honeybee_energy.load.infiltration import Infiltration
    from honeybee_energy.load.ventilation import Ventilation
    from honeybee_energy.lib.schedules import schedule_by_identifier
    from honeybee_energy.lib.programtypes import program_type_by_identifier, \
        building_program_type_by_identifier
    from honeybee_energy.programtype import ProgramType
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# get the always on schedule
always_on = schedule_by_identifier('Always On')


def dup_load(hb_obj, object_name, object_class):
    """Duplicate a load object assigned to a Room or ProgramType."""
    # try to get the load object assgined to the Room or ProgramType
    try:  # assume it's a Room
        load_obj = hb_obj.properties
        for attribute in ('energy', object_name):
            load_obj = getattr(load_obj, attribute)
    except AttributeError:  # it's a ProgramType
        load_obj = getattr(hb_obj, object_name)

    load_id = '{}_{}'.format(hb_obj.identifier, object_name)
    try:  # duplicate the load object
        dup_load = load_obj.duplicate()
        dup_load.identifier = load_id
        return dup_load
    except AttributeError:  # create a new object
        try:  # assume it's People, Lighting, Equipment or Infiltration
            return object_class(load_id, 0, always_on)
        except:  # it's a Ventilation object
            return object_class(load_id)


def assign_load(hb_obj, load_obj, object_name):
    """Assign a load object to a Room or a ProgramType."""
    try:  # assume it's a Room
        setattr(hb_obj.properties.energy, object_name, load_obj)
    except AttributeError:  # it's a ProgramType
        setattr(hb_obj, object_name, load_obj)


def duplicate_and_id_program(program):
    """Duplicate a program and give it a new unique ID."""
    new_prog = program.duplicate()
    new_prog.identifier = '{}_{}'.format(program.identifier, str(uuid.uuid4())[:8])
    return new_prog


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    mod_obj = []
    for obj in _room_or_program:
        if isinstance(obj, Room):
            mod_obj.append(obj.duplicate())
        elif isinstance(obj, ProgramType):
            mod_obj.append(duplicate_and_id_program(obj))
        elif isinstance(obj, str):
            try:
                program = building_program_type_by_identifier(obj)
            except ValueError:
                program = program_type_by_identifier(obj)
            mod_obj.append(duplicate_and_id_program(program))
        else:
            raise TypeError('Expected Honeybee Room or ProgramType. '
                            'Got {}.'.format(type(obj)))

    # assign the people_per_floor_
    if len(people_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            people = dup_load(obj, 'people', People)
            people.people_per_area = longest_list(people_per_floor_, i)
            assign_load(obj, people, 'people')

    # assign the lighting_per_floor_
    if len(lighting_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            lighting = dup_load(obj, 'lighting', Lighting)
            lighting.watts_per_area = longest_list(lighting_per_floor_, i)
            assign_load(obj, lighting, 'lighting')

    # assign the electric_per_floor_
    if len(electric_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            equip = dup_load(obj, 'electric_equipment', ElectricEquipment)
            equip.watts_per_area = longest_list(electric_per_floor_, i)
            assign_load(obj, equip, 'electric_equipment')

    # assign the gas_per_floor_
    if len(gas_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            equip = dup_load(obj, 'gas_equipment', GasEquipment)
            equip.watts_per_area = longest_list(gas_per_floor_, i)
            assign_load(obj, equip, 'gas_equipment')

    # assign the hot_wtr_per_floor_
    if len(hot_wtr_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            shw = dup_load(obj, 'service_hot_water', ServiceHotWater)
            shw.flow_per_area = longest_list(hot_wtr_per_floor_, i)
            assign_load(obj, shw, 'service_hot_water')

    # assign the infilt_per_exterior_
    if len(infilt_per_exterior_) != 0:
        for i, obj in enumerate(mod_obj):
            infilt = dup_load(obj, 'infiltration', Infiltration)
            infilt.flow_per_exterior_area = longest_list(infilt_per_exterior_, i)
            assign_load(obj, infilt, 'infiltration')

    # assign the vent_per_floor_
    if len(vent_per_floor_) != 0:
        for i, obj in enumerate(mod_obj):
            vent = dup_load(obj, 'ventilation', Ventilation)
            vent.flow_per_area = longest_list(vent_per_floor_, i)
            assign_load(obj, vent, 'ventilation')

    # assign the vent_per_person_
    if len(vent_per_person_) != 0:
        for i, obj in enumerate(mod_obj):
            vent = dup_load(obj, 'ventilation', Ventilation)
            vent.flow_per_person = longest_list(vent_per_person_, i)
            assign_load(obj, vent, 'ventilation')

    # assign the vent_ach_
    if len(vent_ach_) != 0:
        for i, obj in enumerate(mod_obj):
            vent = dup_load(obj, 'ventilation', Ventilation)
            vent.air_changes_per_hour = longest_list(vent_ach_, i)
            assign_load(obj, vent, 'ventilation')



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct an Equipment object into its constituient properties.
-

    Args:
        _equip: An ElectricEquipment or a GasEquipment object to be deconstructed.
    
    Returns:
        name_: Text string for the equipment display name.
         watts_per_area: A numerical value for the equipment power density in
            Watts per square meter of floor area.
        schedule: A fractional for the use of equipment over the course of the year.
            The fractional values will get multiplied by the watts_per_area to
            yield a complete equipment profile.
        radiant_fract: A number between 0 and 1 for the fraction of the total
            equipment load given off as long wave radiant heat.
        latent_fract: A number between 0 and 1 for the fraction of the total
            equipment load that is latent (as opposed to sensible).
        lost_fract: A number between 0 and 1 for the fraction of the total
            equipment load that is lost outside of the zone and the HVAC system.
            Typically, this is used to represent heat that is exhausted directly
            out of a zone (as you would for a stove).
        is_gas: Will be True if the input Equipment object is for GasEquipment;
            False if it is for ElectricEquipment.
"""

ghenv.Component.Name = "HB Deconstruct Equipment"
ghenv.Component.NickName = 'DecnstrEquipment'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.equipment import ElectricEquipment, GasEquipment
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _equip is not None:
    # check the input
    assert isinstance(_equip, (ElectricEquipment, GasEquipment)), \
        'Expected Equipment object. Got {}.'.format(type(_equip))

    # get the properties of the object
    name = _equip.display_name
    watts_per_area = _equip.watts_per_area
    schedule = _equip.schedule
    radiant_fract = _equip.radiant_fraction
    latent_fract = _equip.latent_fraction
    lost_fract = _equip.lost_fraction
    is_gas = isinstance(_equip, GasEquipment)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a simple window material to describe an entire glazing system, including
glass, gaps, and frame. This material can be plugged into the "HB Window
Construction" component.
-

    Args:
        _name: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _u_factor: A number for the U-factor of the glazing system [W/m2-K]
            including standard air gap resistances on either side of the
            glazing system.
        _shgc: A number between 0 and 1 for the solar heat gain coefficient
            of the glazing system. This includes both directly transmitted solar
            heat as well as solar heat that is absorbed by the glazing system and
            conducts towards the interior.
        _t_vis_: A number between 0 and 1 for the visible transmittance of the
            glazing system. (Default: 0.6).

    Returns:
        mat: A window material that describes an entire glazing system, including
            glass, gaps, and frame and can be assigned to a Honeybee Window
            construction.
"""

ghenv.Component.Name = 'HB Window Material'
ghenv.Component.NickName = 'WindowMat'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.glazing import EnergyWindowMaterialSimpleGlazSys
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _t_vis_ = 0.6 if _t_vis_ is None else _t_vis_
    name = clean_and_id_ep_string('WindowMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    mat = EnergyWindowMaterialSimpleGlazSys(name, _u_factor, _shgc, _t_vis_)
    if _name_ is not None:
        mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Set any Faces of Rooms with missing adjacencies to Adiabatic.
_
This is useful when simulating a subset of Rooms from a larger Model.
_
If any of the Faces with missing adjacencies have sub-faces, these will be removed
in order to accommodate the adiabatic condition. Similarly, if the Face is an
AirBoundary, the type will be set to a Wall.
-

    Args:
        _rooms: A list of Honeybee Rooms which will have its adjacencies patched
            with Adiabatic boundary conditions. This can also be an entire
            honyebee Model. Any adjacnecy not found across all of the rooms
            will be replaced with an Adiabatic boundary.

    Returns:
        rooms: Rooms that have had their missing adjacencies patched.
"""

ghenv.Component.Name = 'HB Patch Missing Adjacency'
ghenv.Component.NickName = 'PatchAdj'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))


try:
    from ladybug_rhino.grasshopper import all_required_inputs
    from ladybug_rhino.config import tolerance, units_system
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # collect all rooms and duplicate them
    rooms = []
    for hb_obj in _rooms:
        if isinstance(hb_obj, Model):
            rooms.extend(hb_obj.rooms)
        elif isinstance(hb_obj, Room):
            rooms.append(hb_obj)
        else:
            raise ValueError('Expected Room or Model object. Got {}.'.format(type(hb_obj)))
    rooms = [room.duplicate() for room in rooms]  # duplicate to avoid editing input

    # patch adjacency across all of the Rooms
    adj_model = Model('patch_adj_model', rooms=rooms, tolerance=tolerance,
                      units=units_system())
    adj_model.properties.energy.missing_adjacencies_to_adiabatic()
    rooms = adj_model.rooms



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply ProgramType objects to Rooms.
-

    Args:
        _rooms: Honeybee Rooms to which the input load objects should be assigned.
        _program: A ProgramType object to apply to the input rooms,
    
    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with their loads edited.
"""

ghenv.Component.Name = "HB Apply ProgramType"
ghenv.Component.NickName = 'ApplyProgram'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "1"

try:
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from honeybee_energy.lib.programtypes import program_type_by_identifier, \
        building_program_type_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    rooms = [obj.duplicate() for obj in _rooms if isinstance(obj, Room)]

    # apply the program to the rooms
    for i, room in enumerate(rooms):
        prog = longest_list(_program, i)
        if isinstance(prog, str):  # get the program object if it is a string
            try:
                prog = building_program_type_by_identifier(prog)
            except ValueError:
                prog = program_type_by_identifier(prog)
        room.properties.energy.program_type = prog


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse the peak load and HVAC component sizes from an SQL result file that has
been generated from an energy simulation.
-

    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.
        comp_type_: An optional name of a HVAC component type, which will filter
            the HVAC components that appear in the output comp_props and
            comp_values. Connecting nothing here will mean that all HVAC
            component sizes are imported and a full list of possible components
            will appear in the comp_types output.

    Returns:
        zone_names: A list of zone names (honeybee Room identifiers) that correspond
            to the zone_peak_load and zone_peak_heat below.
        zone_peak_cool: A list of numbers for the peak cooling load of each zone
            on the summer design day. These correspond to the zone_names above.
            All values are in Watts.
        zone_peak_heat: A list of numbers for the peak heating load of each zone
            on the winter design day. These correspond to the zone_names above.
            All values are in Watts.
        comp_types: A list of HVAC component types that are available in the results.
            This will be equal to the input comp_type_ if a value is connected.
        comp_properties: A list of text descriptions for HVAC component properties.
            These correspond to the comp_values below.
        comp_values: Values denoting the size of various zone HVAC components 
            (eg. zone terminal sizes, boiler/chiller sizes, lengths of chilled
            beams, etc.). These correspond to the comp_properties above.
            Values may be in different units (eg. W, m/s, etc.) depending
            on the property of the component being observed.
"""

ghenv.Component.Name = 'HB Read HVAC Sizing'
ghenv.Component.NickName = 'ReadHVAC'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess
import json

try:
    from ladybug.sql import SQLiteResult, ZoneSize, ComponentSize
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # create the lists to be filled
    zone_names = []
    zone_peak_cool = []
    zone_peak_heat = []
    comp_properties_mtx = []
    comp_values_mtx = []

    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(_sql)  # create the SQL result parsing object
        zone_cooling_sizes = sql_obj.zone_cooling_sizes
        zone_heating_sizes = sql_obj.zone_heating_sizes
        if comp_type_ is None:
            comp_types = sql_obj.component_types
            component_sizes = sql_obj.component_sizes
        else:
            comp_types = comp_type_
            component_sizes = sql_obj.component_sizes_by_type(comp_type_)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the zone sizes via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'zone-sizes', _sql]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        zone_size_dicts = json.loads(stdout[0])
        zone_cooling_sizes = [ZoneSize.from_dict(zs) for zs in zone_size_dicts['cooling']]
        zone_heating_sizes = [ZoneSize.from_dict(zs) for zs in zone_size_dicts['heating']]
        # Execute the honybee CLI to obtain the component sizes via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'component-sizes', _sql]
        if comp_type_ is not None:
            comp_types = comp_type_
            cmds.extend(['--component-type', comp_type_])
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        comp_size_dicts = json.loads(stdout[0])
        component_sizes = [ComponentSize.from_dict(cs) for cs in comp_size_dicts]
        if comp_type_ is None:  # get a set of all unique component types
            _comp_types = set()
            for comp in component_sizes:
                _comp_types.add(comp.component_type)
            comp_types = list(_comp_types)

    # get the peak zone heating and cooling from the ZoneSize objects
    for zone_size in zone_cooling_sizes:
        zone_names.append(zone_size.zone_name)
        zone_peak_cool.append(zone_size.calculated_design_load)
    for zone_size in zone_heating_sizes:
        zone_peak_heat.append(zone_size.calculated_design_load)

    # get the HVAC component sizes from the ComponentSize objects
    for comp_size in component_sizes:
        comp_properties_mtx.append(comp_size.descriptions)
        comp_values_mtx.append(comp_size.values)
    # convert HVAC components to data trees
    comp_properties = list_to_data_tree(comp_properties_mtx)
    comp_values = list_to_data_tree(comp_values_mtx)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Adjust the properties of a Radiant HVAC that has been assigned to Honeybee Rooms.
_
Because Radiant HVAC systems interact with the conditioned rooms through the
thermal mass of the constructions in which they are embedded, their design
often requires
-

    Args:
        _rooms: Honeybee Rooms that have a Radiant HVAC assigned to them, which are to
            have their radiant properties adjusted. This can also be a Honeybee
            Model for which all Rooms with a Radiant HVAC sill be adjusted.
        radiant_type_: Text to indicate which faces are thermally active by default.
            Note that systems are assumed to be embedded in concrete slabs
            with no insulation within the slab unless otherwise specified.
            Choose from the following. (Default: Floor).
                * Floor
                * Ceiling
                * FloorWithCarpet
                * CeilingMetalPanel
                * FloorWithHardwood
        min_op_time_: A number for the minimum number of hours of operation
            for the radiant system before it shuts off. (Default: 1).
        switch_time_: A number for the minimum number of hours for when the system
            can switch between heating and cooling. (Default: 24).

    Returns:
        rooms: The input Rooms with the radiant HVAC properties edited.
"""

ghenv.Component.Name = 'HB Radiant HVAC Properties'
ghenv.Component.NickName = 'RadiantHVAC'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the honeybee extension
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.hvac.doas.radiant import RadiantwithDOAS
    from honeybee_energy.hvac.heatcool.radiant import Radiant
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and duplicate the rooms
    rooms = []
    for hb_obj in _rooms:
        if isinstance(hb_obj, Model):
            rooms.extend([room.duplicate() for room in hb_obj.rooms])
        elif isinstance(hb_obj, Room):
            rooms.append(hb_obj.duplicate())
        else:
            raise ValueError(
                'Expected Honeybee Room or Model. Got {}.'.format(type(hb_obj)))

    # collect all of the rooms with radiant systems assigned to them
    rad_hvac_dict = {}
    for room in rooms:
        r_hvac = room.properties.energy.hvac
        if isinstance(r_hvac, (RadiantwithDOAS, Radiant)):
            try:
                rad_hvac_dict[r_hvac.identifier].append(room)
            except KeyError:
                rad_hvac_dict[r_hvac.identifier] = [room]

    # adjust the properties of each radiant HVAC that was found
    for r_hvac_rooms in rad_hvac_dict.values():
        new_r_hvac = r_hvac_rooms[0].properties.energy.hvac.duplicate()
        if min_op_time_:
            new_r_hvac.minimum_operation_time = min_op_time_
        if switch_time_:
            new_r_hvac.switch_over_time = switch_time_
        if radiant_type_:
            new_r_hvac.radiant_type = radiant_type_
        for new_r in r_hvac_rooms:
            new_r.properties.energy.hvac = new_r_hvac

    # raise a warning if no rooms with a radiant HVAC were found
    if len(rad_hvac_dict) == 0:
        msg = 'No Rooms with a Radiant HVAC were found among the connected _rooms.\n' \
            'Make sure that a Radiant HVAC has been assigned with either the\n' \
            '"HB DOAS HVAC" or "HB HeatCool HVAC" component.'
        print(msg)
        give_warning(ghenv.Component, msg)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Deconstruct a material into its constituient attributes and values.
-

    Args:
        _mat: A material to be deconstructed. This can also be text for a
            material to be looked up in the material library.

    Returns:
        values: List of values for the attributes that define the material.
        attr_names: List of text that is the same length as the values, which
            notes the attribute name for each value.
        r_val_si: R-value of the material in m2-K/W. Note that R-values do NOT
            include the resistance of air films on either side of the material.
        r_val_ip: R-value of the material in h-ft2-F/Btu. Note that R-values do NOT
            include the resistance of air films on either side of the material.
"""

ghenv.Component.Name = "HB Deconstruct Material"
ghenv.Component.NickName = 'DecnstrMat'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "2"

import re

try:  # import the honeybee-energy dependencies
    from honeybee_energy.reader import parse_idf_string
    from honeybee_energy.lib.materials import opaque_material_by_identifier
    from honeybee_energy.lib.materials import window_material_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
try:
    from ladybug.datatype.rvalue import RValue
    from ladybug.datatype.uvalue import UValue
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_mat, str):
        try:
            _mat = opaque_material_by_identifier(_mat)
        except ValueError:
            _mat = window_material_by_identifier(_mat)

    # get the values and attribute names
    mat_str = str(_mat)
    values = parse_idf_string(mat_str)
    attr_name_pattern1 = re.compile(r'!- (.*)\n')
    attr_name_pattern2 = re.compile(r'!- (.*)$')
    attr_names = attr_name_pattern1.findall(mat_str) + \
        attr_name_pattern2.findall(mat_str)

    # get the r-value
    try:
        r_val_si = _mat.r_value
        r_val_ip = RValue().to_ip([r_val_si], 'm2-K/W')[0][0]

        # give a warning if there's a negative R-value for a vertical surface
        if r_val_si <= 0:
            msg = 'Material "{}" has an overall R-value that is less than the\n' \
                'resistance of vertically-oriented air films. This indicates that the ' \
                'construction is only suitable for horizontal/skylight geometry.'.format(
                    _mat.display_name)
            print(msg)
            give_warning(ghenv.Component, msg)
    except AttributeError:
        r_val_si = 'varies'
        r_val_ip = 'varies'

    # re-order the E+ attributes of opaque constructions to align with component
    if len(attr_names) == 9:
        values.insert(5, values.pop(1))
        attr_names.insert(5, attr_names.pop(1))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create an EnergyPlus shade construction. Note that Shade objects in EnergyPlus
do not have layers and are only defined by their exterior reflectance.
-

    Args:
        _name: Text to set the name for the Construction and to be incorporated
            into a unique Construction identifier.
        _sol_ref_: A number between 0 and 1 for the solar reflectance
            of the construction. Default: 0.2.
        _vis_ref_: A number between 0 and 1 for the visible reflectance
            of the construction. Default: 0.2.
        specular_: A boolean to note whether the reflection off the shade
            should be diffuse (False) or specular (True). Set to True if the
            construction is representing a glass facade or a mirror
            material. Default: False.

    Returns:
        constr: A shade construction that can be assigned to Honeybee
            Shades or ConstructionSets.
"""

ghenv.Component.Name = "HB Shade Construction"
ghenv.Component.NickName = 'ShadeConstr'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "4"

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.construction.shade import ShadeConstruction
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

if all_required_inputs(ghenv.Component):
    # set default values
    _sol_ref_ = 0.2 if _sol_ref_ is None else _sol_ref_
    _vis_ref_ = 0.2 if _vis_ref_ is None else _vis_ref_
    specular_ = False if specular_ is None else specular_
    name = clean_and_id_ep_string('ShadeConstruction') if _name_ is None else \
        clean_ep_string(_name_)

    # create the construction
    constr = ShadeConstruction(name, _sol_ref_, _vis_ref_, specular_)
    if _name_ is not None:
        constr.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct an Infiltration object into its constituient properties.
-

    Args:
        _infil: An Infiltration object to be deconstructed.
    
    Returns:
        name: Text string for the infiltration display name.
        flow_per_ext_area: A numerical value for the intensity of infiltration
            in m3/s per square meter of exterior surface area. Typical values for
            this property are as follows (note all values are at typical building
            pressures of ~4 Pa):
                * 0.0001 (m3/s per m2 facade) - Tight building
                * 0.0003 (m3/s per m2 facade) - Average building
                * 0.0006 (m3/s per m2 facade) - Leaky building
        schedule: A fractional schedule for the infiltration over the course
            of the year. The fractional values will get multiplied by the
            flow_per_exterior_area to yield a complete infiltration profile.
"""

ghenv.Component.Name = "HB Deconstruct Infiltration"
ghenv.Component.NickName = 'DecnstrInfiltration'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.infiltration import Infiltration
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _infil is not None:
    # check the input
    assert isinstance(_infil, Infiltration), \
        'Expected Infiltration object. Got {}.'.format(type(_infil))
    
    # get the properties of the object
    name = _infil.display_name
    flow_per_ext_area = _infil.flow_per_exterior_area
    schedule = _infil.schedule



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Separate data collections of energy simulation results by object and face type.
Input data must be for Faces, Apertures, Doors, or any combination of these objects.
_
This component can also be used to normalize such data by area.
-

    Args:
        _data: A list of data collections output from an energy simulation, which
            will be separated by object and face type. Data collections can be
            of any class (eg. MonthlyCollection, DailyCollection) but they
            should all have headers with metadata dictionaries with 'Surface'
            keys. These keys will be used to match the data in the collections
            to the input faces.
        _hb_objs: An array of honeybee Rooms, Faces, Apertures or Doors, which will
            be matched with the _data. This can also be an entire Model.
        norm_: Boolean to note whether results should be normalized by the face/sub-face
            area if the data type of the data_colections supports it. (Default: False)

    Returns:
        walls: Data collections with results for Walls with an Outdoors or Ground
            boundary condition.
        interior_walls: Data collections with results for Walls with a Surface or
            Adiabatic boundary condition.
        roofs: Data collections with results for RoofCeilings with an Outdoors or
            Ground boundary condition.
        ceilings: Data collections with results for RoofCeilings with a Surface
            or Adiabatic boundary condition.
        exterior_floors: Data collections with results for Floors with an Outdoors
            or Ground boundary condition.
        interior_floors: Data collections with results for Floors with a Surface
            or Adiabatic boundary condition.
        apertures: Data collections with results for Apertures with an Outdoors
            boundary condition.
        interior_apertures: Data collections with results for Apertures with a
            Surface boundary condition.
        doors: Data collections with results for Doors with an Outdoors boundary
            condition.
        interior_doors: Data collections with results for Doors with a Surface
            boundary condition.
"""

ghenv.Component.Name = 'HB Face Result by Type'
ghenv.Component.NickName = 'FaceResultByType'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from honeybee.model import Model
    from honeybee.room import Room
    from honeybee.face import Face
    from honeybee.aperture import Aperture
    from honeybee.door import Door
    from honeybee.boundarycondition import Surface, Adiabatic
    from honeybee.facetype import Wall, RoofCeiling, Floor, AirBoundary
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.match import match_faces_to_data
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
    from ladybug_rhino.config import conversion_to_meters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def matched_areas(matched_objects):
    """Get an iterator of the areas from a list of matched objects."""
    flat_faces = (obj[0] for obj in matched_objects)
    flat_geo = (face.geometry if not isinstance(face, Face)
                else face.punched_geometry for face in flat_faces)
    return (face.area for face in flat_geo)


if all_required_inputs(ghenv.Component):
    # extract any faces from input Rooms or Models and convert geo to meters
    faces = []
    for hb_obj in _hb_objs:
        if isinstance(hb_obj, Model):
            for room in hb_obj.rooms:
                faces.extend(room.faces)
        elif isinstance(hb_obj, Room):
            faces.extend(hb_obj.faces)
        else:
            faces.append(hb_obj)
    m_convert = conversion_to_meters()
    if m_convert != 1:  # duplicate and scale all objects to meters
        faces = [face.duplicate() for face in faces]
        [face.scale(m_convert) for face in faces]

    # match the data with the faces
    matched_tups = match_faces_to_data(_data, faces)

    # normalize the data if requested
    if norm_:
        norm_types = [tup[1].header.data_type.normalized_type for tup in matched_tups]
        new_tups = []
        for area, nt, tup in zip(matched_areas(matched_tups), norm_types, matched_tups):
            if nt is None:  # data is not normalizable
                new_tups.append(tup)
            else:
                new_tups.append((tup[0], tup[1].normalize_by_area(area, 'm2')))
        matched_tups = new_tups

    # lists to be filled with data collections
    walls = []
    interior_walls = []
    roofs = []
    ceilings = []
    exterior_floors = []
    interior_floors = []
    apertures = []
    interior_apertures = []
    doors = []
    interior_doors = []

    # loop through the tuples and sort them by object and face type
    for tup in matched_tups:
        obj, data = tup
        bc = obj.boundary_condition
        if isinstance(obj, Face):
            if isinstance(obj.type, Wall):
                if isinstance(bc, (Surface, Adiabatic)):
                    interior_walls.append(data)
                else:
                    walls.append(data)
            elif isinstance(obj.type, RoofCeiling):
                if isinstance(bc, (Surface, Adiabatic)):
                    ceilings.append(data)
                else:
                    roofs.append(data)
            elif isinstance(obj.type, Floor):
                if isinstance(bc, (Surface, Adiabatic)):
                    interior_floors.append(data)
                else:
                    exterior_floors.append(data)
        elif isinstance(obj, Aperture):
            if isinstance(bc, Surface):
                interior_apertures.append(data)
            else:
                apertures.append(data)
        elif isinstance(obj, Door):
            if isinstance(bc, Surface):
                interior_doors.append(data)
            else:
                doors.append(data)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a schedule from a single constant value or a list of 24 hourly values
repeating continuously over every day of the year.
-

    Args:
        _values: A list of 24 values that represent the schedule values at each
            hour of the day. This can also be a single constant value for the
            whole day.
        _name_: Text to set the name for the Schedule and to be incorporated
            into a unique Schedule identifier.
        _type_limit_: A text string from the identifier of the ScheduleTypeLimit to
            be looked up in the schedule type limit library. This can also be a
            custom ScheduleTypeLimit object from the "HB Type Limit" component.
            The input here will be used to validate schedule values against
            upper/lower limits and assign units to the schedule values. Default:
            "Fractional" for values that range continuously between 0 and 1.
            Choose from the following built-in options:
                * Fractional
                * On-Off
                * Temperature
                * Activity Level
                * Power
                * Humidity
                * Angle
                * Delta Temperature
                * Control Level

    Returns:
        report: Reports, errors, warnings, etc.
        schedule: A ScheduleRuleset object that can be assigned to a Room, a Load
            object, or a ProgramType object.
        idf_text: Text strings for the EnergyPlus Schedule that will ultimately
            be written into the IDF for simulation. This can also be used to add
            the schedule to the schedule library that is loaded up upon the start
            of Honeybee by copying this text into the honeybee/library/schedules/
            user_library.idf file.
"""

ghenv.Component.Name = 'HB Constant Schedule'
ghenv.Component.NickName = 'ConstantSchedule'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the ScheduleTypeLimit object
    if _type_limit_ is None:
        _type_limit_ = schedule_type_limit_by_identifier('Fractional')
    elif isinstance(_type_limit_, str):
        _type_limit_ = schedule_type_limit_by_identifier(_type_limit_)

    # create the schedule object
    name = clean_and_id_ep_string('ConstantSchedule') if _name_ is None else \
        clean_ep_string(_name_)
    if len(_values) == 1:
        schedule = ScheduleRuleset.from_constant_value(name, _values[0], _type_limit_)
        idf_text, constant_none = schedule.to_idf()
    else:
        schedule = ScheduleRuleset.from_daily_values(name, _values, timestep=1,
            schedule_type_limit=_type_limit_)
        idf_year, idf_week = schedule.to_idf()
        idf_days = [day_sch.to_idf(_type_limit_) for day_sch in schedule.day_schedules]
        idf_text = [idf_year] + idf_week + idf_days if idf_week is not None \
            else idf_year
    if _name_ is not None:
        schedule.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Search for available ConstructionSets within the honeybee energy standards library.
-

    Args:
        keywords_: Optional keywords to be used to narrow down the output list of
            construction sets. If nothing is input here, all available
            construction sets will be output.
        join_words_: If False or None, this component will automatically split
            any strings of multiple keywords (spearated by spaces) into separate
            keywords for searching. This results in a greater liklihood of
            finding an item in the search but it may not be appropropriate for
            all cases. You may want to set it to True when you are searching for
            a specific phrase that includes spaces. (Default: False).
    
    Returns:
        constr_sets: A list of ConstructionSet identifiers that can be applied
            to Honeybee Rooms.
"""

ghenv.Component.Name = 'HB Search Construction Sets'
ghenv.Component.NickName = 'SearchConstrSet'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the honeybee-core dependencies
    from honeybee.search import filter_array_by_keywords
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.lib.constructionsets import CONSTRUCTION_SETS
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


if len(keywords_) == 0:
    constr_sets = sorted(CONSTRUCTION_SETS)
else:
    split_words = True if join_words_ is None else not join_words_
    constr_sets = sorted(filter_array_by_keywords(
        CONSTRUCTION_SETS, keywords_, split_words))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Deconstruct a ScheduleRuleset into an array of day-long ladybug DataCollections
representing each unique ScheduleDay that defines the ScheduleRuleset.
_
These DataCollections can be used to make visualizations of timeseries schedule
values over each unique day of the schedule using a component like the
"LB Line Chart".
-

    Args:
        _schedule: A ScheduleRuleSet to be deconstructed into DataCollections
            of timeseries schedule values for each unique day. This can also
            be the identifier of a Schedule to be looked up in the schedule library.
        _timestep_: An integer for the number of steps per hour at which to make
            the resulting daily DataCollections.

    Returns:
        day_names: A list of display names for each unique ScheduleDay that
            defines the input ScheduleRuleset.
        day_data: A list of day-long ladybug DataCollections representing each
            unique ScheduleDay that defines the ScheduleRuleset. These can be
            used to make visualizations of timeseries schedule values over each
            day of the schedule using a component like the "LB Line Chart".
        type_limit: The ScheduleTypeLimit object assigned to the ScheduleRuleset.
"""

ghenv.Component.Name = "HB Deconstruct Schedule"
ghenv.Component.NickName = 'DeconstrSch'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = "2"


try:  # import the honeybee-energy dependencies
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    from honeybee_energy.schedule.fixedinterval import ScheduleFixedInterval
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:  # import the ladybug dependencies
    from ladybug.dt import Date
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the schedue from the library if it's a string
    if isinstance(_schedule, str):
        _schedule = schedule_by_identifier(_schedule)

    # process the timestep
    _timestep_ = 1 if _timestep_ is None else _timestep_
    st_date = Date(1, 1)
    type_limit = _schedule.schedule_type_limit

    if isinstance(_schedule, ScheduleRuleset):
        # process the default ScheduleDay
        default = _schedule.default_day_schedule.data_collection(
            st_date, type_limit, _timestep_)
        default.header.metadata['applied'] = 'default'

        # create the ScheduleDay DataCollections
        day_names = [_schedule.default_day_schedule.display_name]
        day_data = [default]
        for rule in _schedule.schedule_rules:
            data_col = rule.schedule_day.data_collection(
                st_date, type_limit, _timestep_)
            meta_str = ', '.join(rule.days_applied) + \
                ' | {} to {}'.format(rule.start_date, rule.end_date)
            data_col.header.metadata['applied'] = meta_str
            day_names.append(rule.schedule_day.display_name)
            day_data.append(data_col)

        # process the design day schedules
        if _schedule.summer_designday_schedule is not None:
            summer = _schedule.summer_designday_schedule.data_collection(
                st_date, type_limit, _timestep_)
            summer.header.metadata['applied'] = 'summer design'
            day_names.append(_schedule.summer_designday_schedule.display_name)
            day_data.append(summer)
        if _schedule.winter_designday_schedule is not None:
            winter = _schedule.winter_designday_schedule.data_collection(
                st_date, type_limit, _timestep_)
            winter.header.metadata['applied'] = 'winter design'
            day_names.append(_schedule.winter_designday_schedule.display_name)
            day_data.append(winter)

    elif isinstance(_schedule, ScheduleFixedInterval):
        msg = 'Input schedule "{}" is a FixedInterval schedule\nand cannot ' \
            'be deconstructed into individual days.\nUse the "HB Schedule to ' \
            'Data" component instead.'.format(_schedule.display_name)
        print(msg)
        give_warning(ghenv.Component, msg)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate a building (or zone) balance temperature from a ideal air load simulation
results. The balance point is the outdoor temperature at which a building switches
between heating and cooling.
_
If the outdoor temperture drops below the balance temperature, the building will
usually be in heating mode and, if the outdoor temperture is above the balance
temperature, the building will usually be in cooling mode.
_
The balance temperture concept is useful for setting things such as automated
blinds and airflow schedules since having these things directly controlled by
hourly cooling or heating demand isn't always straightforward.
_
This component works by taking the average combined heating/cooling values for
each day and the average outdoor air temperature for each day. The days with
the smallest combined heating + cooling will have their daily mean outdoor air
tempertures averaged to produce the output balance temperture.

-
    Args:
        _cooling: Data collection for the annual hourly or daily ideal air cooling
            load output from the "HB Read Room Energy Result" component.  This
            can be for a single room or the entire model.
        _heating: Data collection for the annual hourly or daily ideal air heating
            load output from the "HB Read Room Energy Result" component.  This
            can be for a single room or the entire model.
        _temperature: Data collection for the annual hourly or daily outdoor temperature.
            Most of the time, this should be the dry bulb temperature from the
            "LB Import EPW" component. However other types of temperature like
            sky temperature may improve accuracy since they include the effects
            of solar gain. Note that, whatever type of temperature is plugged
            in here determines the type of balance temperature that is output.
        _day_count_: An integer for the number of days with a low thermal energy
            load that will be averaged together to yield the balance point.
            The use of multiple days is done to help avoid anomalies introduced
            by things like variations between weekday and weekend shcedules.
            It is recommended that this be increased for models with particularly
            high variation in schedules. (Default: 10).

    Returns:
        bal_day_load: The average thermal load on the balance day. This number should
            be close to 0 if the balance temperature is accurate and this output
            is meant to give a sense of the accuracy of the balance temperature value.
        bal_temp: The outdoor balance temperature.
"""

ghenv.Component.Name = 'HB Balance Temperature'
ghenv.Component.NickName = 'BalTemp'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import HourlyContinuousCollection, DailyCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def sum_collection(data):
    """Sum a list of data collections into a single daily collection."""
    total_data = data[0]
    for data_i in data[1:]:
        total_data = total_data + data_i
    if isinstance(total_data, HourlyContinuousCollection):
        total_data = total_data.average_daily()
    elif not isinstance(total_data, DailyCollection):
        raise TypeError('Expected Hourly or Daily data collection. '
                        'Got {}'.format(type(_temperature)))
    return total_data


if all_required_inputs(ghenv.Component):
    # set the defaut day count and ensure temperature is daily
    _day_count_ = 10 if _day_count_ is None else _day_count_
    if isinstance(_temperature, HourlyContinuousCollection):
        _temperature = _temperature.average_daily()
    elif not isinstance(_temperature, DailyCollection):
        raise TypeError('Expected Hourly or Daily data collection for '
                        '_temperature. Got {}'.format(type(_temperature)))

    # convert the heating and cooling collections to a single daily balance
    total_cool = sum_collection(_cooling)
    total_heat = sum_collection(_heating)
    total_load = total_cool + total_heat

    # check that all data collections are annual
    assert len(total_load) >= 365, 'Cooling and heating loads are not annual.'
    assert len(_temperature) >= 365, 'Temperature is not annual.'

    # sort the load to find the days with the lowest load
    temp_sort = []
    load_sort = []
    for i, (load, temp) in enumerate(sorted(zip(total_load, _temperature))):
        if i == _day_count_:
            break
        temp_sort.append(temp)
        load_sort.append(load)

    # return the average temperature and balance day load
    bal_day_load = sum(load_sort) / _day_count_
    bal_temp = sum(temp_sort) / _day_count_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a material for a shade layer in a window construction (like a roller shade).
This material can be plugged into the "HB Window Construction" component.
_
Reflectance and emissivity properties are assumed to be the same on both sides of
the shade. Shades are considered to be perfect diffusers.
-

    Args:
        _name: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _thickness: Number for the thickness of the shade layer in meters.
        _transmittance_:  Number between 0 and 1 for the transmittance of both solar
            radiation and visible light through the shade. (Default: 0.4, which
            is typical of a white diffusing shade).
        _reflectance_: Number between 0 and 1 for the reflectance of both solar
            radiation and visible light off of the shade. (Default: 0.5,
            which is typical of a white diffusing shade).
        _t_infrared_: Long-wave hemisperical transmittance of the shade. (Default: 0).
        _emissivity_: Number between 0 and 1 for the infrared hemispherical
            emissivity of the shade. (Default: 0.9, which is typical of most
            diffusing shade materials).
        _conductivity_: Number for the thermal conductivity of the shade in
            W/m-K. (Default: 0.05, typical of cotton shades).
        _dist_to_glass_: A number between 0.001 and 1.0 for the distance between the
            shade and neighboring glass layers [m]. (Default: 0.05 (50 mm)).
        _open_mult_: Factor between 0 and 1 that is multiplied by the area at the top,
            bottom and sides of the shade for air flow calculations. (Default: 0.5).
        _permeability_: The fraction of the shade surface that is open to air flow.
            Must be between 0 and 0.8. (Default: 0 for no permeability).

    Returns:
        mat: A material for a shade layer in a window construction (like a roller
            shade) that can be assigned to a Honeybee Window construction.
"""

ghenv.Component.Name = 'HB Shade Material'
ghenv.Component.NickName = 'ShadeMat'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.shade import EnergyWindowMaterialShade
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _transmittance_ = 0.4 if _transmittance_ is None else _transmittance_
    _reflectance_ = 0.5 if _reflectance_ is None else _reflectance_
    _t_infrared_ = 0 if _t_infrared_ is None else _t_infrared_
    _emissivity_ = 0.9 if _emissivity_ is None else _emissivity_
    _conductivity_ = 0.9 if _conductivity_ is None else _conductivity_
    _dist_to_glass_ = 0.05 if _dist_to_glass_ is None else _dist_to_glass_
    _open_mult_ = 0.5 if _open_mult_ is None else _open_mult_
    _permeability_ = 0.0 if _permeability_ is None else _permeability_
    name = clean_and_id_ep_string('ShadeMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    mat = EnergyWindowMaterialShade(
        name, _thickness, _transmittance_, _reflectance_,
        _transmittance_, _reflectance_, _t_infrared_, _emissivity_, _conductivity_,
        _dist_to_glass_, _open_mult_, _permeability_)
    if _name_ is not None:
        mat.display_name = _name_



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse an .rdd file from an energy simulation to show all possible outputs that
can be requested from the simulation.
-

    Args:
        _rdd: Full path to a RDD file that was generated by EnergyPlus.
        keywords_: An optional list of keywords that will be used to filter
            the output names.
        join_words_: If False or None, this component will automatically split
            any strings of multiple keywords (spearated by spaces) into separate
            keywords for searching. This results in a greater liklihood of
            finding an item in the search but it may not be appropropriate for
            all cases. You may want to set it to True when you are searching for
            a specific phrase that includes spaces. (Default: False).
    
    Returns:
        outputs: A list of EnergyPlus output names as strings (eg. 'Surface Window
            System Solar Transmittance'). If no keywords are input, this will
            be a list of all possible outputs that can be requested from the
            simulation. Outputs can be requested from the simulation by plugging
            them into the output_names_ of the "HB Custom Simulation Output" component. 
"""

ghenv.Component.Name = 'HB Read Result Dictionary'
ghenv.Component.NickName = 'ReadRDD'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the honeybee-core dependencies
    from honeybee.search import filter_array_by_keywords
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.rdd import RDD
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    rdd_obj = RDD(_rdd)
    if len(keywords_) == 0:
        outputs = rdd_obj.output_names
    else:
        split_words = True if join_words_ is None else not join_words_
        outputs = filter_array_by_keywords(rdd_obj.output_names, keywords_, split_words)
